local N = 100
local pf = require "pathfun"
local pmaps = require "pmaps"
local Vec2 = require("pathfun.master").Vec2

local nav = pf.Navigation(pmaps)
local love = _G["love"]

local love_polygons = {}
for i, t in ipairs(pmaps[1]) do
    local tmp = {}
    love_polygons[i] = tmp
    for _, v in ipairs(t) do
        tmp[#tmp + 1] = v[1]
        tmp[#tmp + 1] = v[2]
    end
end

local points
-- s =1, l=0.5
local function hue(h) --input between 0 and 1
	h = h*6
	local x = 1 - math.abs(h%2 - 1)
	if h < 1 then
        return 1, x, 0
	elseif h < 2 then
        return x, 1, 0
	elseif h < 3 then
        return 0, 1, x
	elseif h < 4 then
        return 0, x, 1
	elseif h < 5 then
        return x, 0, 1
	else
        return 1, 0, x
	end
end

local function initalise()
    points = {}
    for i = 1, N do
        local t = {}
        local x, y = love.math.random(0, love.graphics.getWidth() - 1), love.math.random(0, love.graphics.getHeight() - 1)
        local v = Vec2(x, y)
        local path = nav:_shortest_path(v, v)
        t.pos = path[1]
        t.speed = 50*(1 + love.math.random())
        t.col = {hue((i - 1)/N)}
        points[i] = t
    end
end

function love.load()
    initalise()
end

function love.draw()
    love.graphics.setColor(1,1,1,.2)
    for _, poly in ipairs(love_polygons) do
        love.graphics.polygon("fill", poly)
    end

    for _, P in ipairs(points) do
        love.graphics.setColor(P.col)
        local v = P.pos
        love.graphics.circle("fill", v.x, v.y, 2)
    end
    love.graphics.setColor(1,1,1)

    love.graphics.print(string.format("# of points: %d", N), 20, 580)
    love.graphics.print(string.format("FPS: %d", love.timer.getFPS()), 20, 600)
    love.graphics.print("Left-click anywhere to sent the points to random positions", 20, 620)
    love.graphics.print("Right-click somwhere to send there all the points", 20, 640)
    love.graphics.print("Press up/down to change the number of points", 20, 660)
    love.graphics.print("Points move at different speeds, randomly assigned at the beginning", 20, 680)
end

function love.mousepressed(x, y, button)
    for _, P in ipairs(points) do
        local X, Y
        if button == 1 then
            X, Y = love.math.random(0, love.graphics.getWidth() - 1), love.math.random(0, love.graphics.getHeight() - 1)
        else
            X, Y = x, y
        end
        local v = Vec2(X, Y)
        P.path = nav:_shortest_path(P.pos, v)
        P.piece = 1
        P.t = 0
        P.distances = {}
        for i = 1, #P.path - 1 do
            P.distances[i] = (P.path[i + 1] - P.path[i]):len()
        end
        P.times = {[0]=0}
        for i = 1, #P.path - 1 do
            P.times[i] = P.times[i - 1] + P.distances[i]/P.speed
        end
        P.times[#P.times + 1] = math.huge
        P.moving = true
    end
end

function love.update(dt)
    for _, P in ipairs(points) do
        if P.moving then
            P.t = P.t + dt
            local factor
            for i = P.piece, #P.path do
                if P.t >= P.times[i - 1] and P.t <= P.times[i] then
                    P.piece = i
                    factor = (P.t - P.times[i - 1])/(P.times[i] - P.times[i - 1])
                    break
                end
            end
            if P.piece == #P.path then
                P.pos = P.path[#P.path]
                P.moving = false
            else
                local A, B = P.path[P.piece], P.path[P.piece + 1]
                P.pos = A*(1 - factor) + B*factor
            end
        end
    end
end

function love.keypressed(key)
    if key == "up" then
        N = N + 10
        initalise()
    elseif key == "down" then
        N = math.max(N - 10, 0)
        initalise()
    end
end